// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    value_derive_common::{
        append_field, derive_clone_hash_eq_partialeq_for_struct, get_field_names, get_target,
        make_connect_impl, Bounds, Builder, FieldsKind, ParsedField, ValueDeriveGenerics,
    },
    value_derive_struct::{self, ParsedStruct, ParsedStructNames, StructOptions},
    Errors, HdlAttr,
};
use proc_macro2::TokenStream;
use quote::{format_ident, quote, quote_spanned, ToTokens};
use syn::{
    parse_quote, parse_quote_spanned, punctuated::Punctuated, spanned::Spanned, token::Brace,
    Field, FieldMutability, Fields, FieldsNamed, Generics, Ident, Index, ItemEnum, ItemStruct,
    Member, Path, Token, Type, Variant, Visibility,
};

crate::options! {
    #[options = EnumOptions]
    enum EnumOption {
        OutlineGenerated(outline_generated),
        ConnectInexact(connect_inexact),
        Bounds(where_, Bounds),
        Target(target, Path),
    }
}

crate::options! {
    #[options = VariantOptions]
    enum VariantOption {}
}

crate::options! {
    #[options = FieldOptions]
    enum FieldOption {}
}

enum VariantValue {
    None,
    Direct {
        value_type: Type,
    },
    Struct {
        value_struct: ItemStruct,
        parsed_struct: ParsedStruct,
    },
}

impl VariantValue {
    fn is_none(&self) -> bool {
        matches!(self, Self::None)
    }
    fn value_ty(&self) -> Option<Type> {
        match self {
            VariantValue::None => None,
            VariantValue::Direct { value_type } => Some(value_type.clone()),
            VariantValue::Struct { value_struct, .. } => {
                let (_, type_generics, _) = value_struct.generics.split_for_impl();
                let ident = &value_struct.ident;
                Some(parse_quote! { #ident #type_generics })
            }
        }
    }
}

struct ParsedVariant {
    options: HdlAttr<VariantOptions>,
    ident: Ident,
    fields_kind: FieldsKind,
    fields: Vec<ParsedField<FieldOptions>>,
    value: VariantValue,
}

impl ParsedVariant {
    fn parse(
        errors: &mut Errors,
        variant: Variant,
        enum_options: &EnumOptions,
        enum_vis: &Visibility,
        enum_ident: &Ident,
        enum_generics: &Generics,
    ) -> Self {
        let target = get_target(&enum_options.target, enum_ident);
        let Variant {
            mut attrs,
            ident,
            fields,
            discriminant,
        } = variant;
        if let Some((eq, _)) = discriminant {
            errors.error(eq, "#[derive(Value)]: discriminants not allowed");
        }
        let variant_options = errors
            .unwrap_or_default(HdlAttr::parse_and_take_attr(&mut attrs))
            .unwrap_or_default();
        let (fields_kind, parsed_fields) =
            ParsedField::parse_fields(errors, &mut fields.clone(), true);
        let value = match (&fields_kind, &*parsed_fields) {
            (FieldsKind::Unit, _) => VariantValue::None,
            (
                FieldsKind::Unnamed(_),
                [ParsedField {
                    options,
                    vis: _,
                    name: Member::Unnamed(Index { index: 0, span: _ }),
                    ty,
                }],
            ) => {
                let FieldOptions {} = options.body;
                VariantValue::Direct {
                    value_type: ty.clone(),
                }
            }
            _ => {
                let variant_value_struct_ident =
                    format_ident!("__{}__{}", enum_ident, ident, span = ident.span());
                let variant_type_struct_ident =
                    format_ident!("__{}__{}__Type", enum_ident, ident, span = ident.span());
                let mut value_struct_fields = fields.clone();
                let (_, type_generics, _) = enum_generics.split_for_impl();
                append_field(
                    &mut value_struct_fields,
                    Field {
                        attrs: vec![HdlAttr::from(value_derive_struct::FieldOptions {
                            flip: None,
                            skip: Some(Default::default()),
                        })
                        .to_attr()],
                        vis: enum_vis.clone(),
                        mutability: FieldMutability::None,
                        ident: Some(Ident::new("__phantom", ident.span())),
                        colon_token: None,
                        ty: parse_quote_spanned! {ident.span()=>
                            ::fayalite::__std::marker::PhantomData<#target #type_generics>
                        },
                    },
                );
                let (value_struct_fields_kind, value_struct_parsed_fields) =
                    ParsedField::parse_fields(errors, &mut value_struct_fields, false);
                let value_struct = ItemStruct {
                    attrs: vec![parse_quote! { #[allow(non_camel_case_types)] }],
                    vis: enum_vis.clone(),
                    struct_token: Token![struct](ident.span()),
                    ident: variant_value_struct_ident.clone(),
                    generics: enum_generics.clone(),
                    fields: value_struct_fields,
                    semi_token: None,
                };
                VariantValue::Struct {
                    value_struct,
                    parsed_struct: ParsedStruct {
                        options: StructOptions {
                            outline_generated: None,
                            fixed_type: Some(Default::default()),
                            where_: Some((
                                Default::default(),
                                Default::default(),
                                ValueDeriveGenerics::get(
                                    enum_generics.clone(),
                                    &enum_options.where_,
                                )
                                .fixed_type_generics
                                .where_clause
                                .into(),
                            )),
                            target: None,
                            connect_inexact: enum_options.connect_inexact,
                        }
                        .into(),
                        vis: enum_vis.clone(),
                        struct_token: Default::default(),
                        generics: enum_generics.clone(),
                        fields_kind: value_struct_fields_kind,
                        fields: value_struct_parsed_fields,
                        semi_token: None, // it will fill in the semicolon if needed
                        skip_check_fields: true,
                        names: ParsedStructNames {
                            ident: variant_value_struct_ident.clone(),
                            type_struct_debug_ident: Some(format!("{enum_ident}::{ident}::Type")),
                            type_struct_ident: variant_type_struct_ident,
                            match_variant_ident: None,
                            builder_struct_ident: None,
                            mask_match_variant_ident: None,
                            mask_type_ident: None,
                            mask_type_debug_ident: Some(format!(
                                "AsMask<{enum_ident}::{ident}>::Type"
                            )),
                            mask_value_ident: None,
                            mask_value_debug_ident: Some(format!("AsMask<{enum_ident}::{ident}>")),
                            mask_builder_struct_ident: None,
                        },
                    },
                }
            }
        };
        ParsedVariant {
            options: variant_options,
            ident,
            fields_kind,
            fields: parsed_fields,
            value,
        }
    }
}

struct ParsedEnum {
    options: HdlAttr<EnumOptions>,
    vis: Visibility,
    enum_token: Token![enum],
    ident: Ident,
    generics: Generics,
    brace_token: Brace,
    variants: Vec<ParsedVariant>,
}

impl ParsedEnum {
    fn parse(item: ItemEnum) -> syn::Result<Self> {
        let ItemEnum {
            mut attrs,
            vis,
            enum_token,
            ident,
            generics,
            brace_token,
            variants,
        } = item;
        let mut errors = Errors::new();
        let enum_options = errors
            .unwrap_or_default(HdlAttr::parse_and_take_attr(&mut attrs))
            .unwrap_or_default();
        let variants = variants
            .into_iter()
            .map(|variant| {
                ParsedVariant::parse(
                    &mut errors,
                    variant,
                    &enum_options.body,
                    &vis,
                    &ident,
                    &generics,
                )
            })
            .collect();
        errors.finish()?;
        Ok(ParsedEnum {
            options: enum_options,
            vis,
            enum_token,
            ident,
            generics,
            brace_token,
            variants,
        })
    }
}

impl ToTokens for ParsedEnum {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            options,
            vis,
            enum_token,
            ident: enum_ident,
            generics: enum_generics,
            brace_token,
            variants,
        } = self;
        let EnumOptions {
            outline_generated: _,
            connect_inexact,
            where_,
            target,
        } = &options.body;
        let target = get_target(target, enum_ident);
        let ValueDeriveGenerics {
            generics: _,
            fixed_type_generics,
        } = ValueDeriveGenerics::get(enum_generics.clone(), where_);
        let (fixed_type_impl_generics, fixed_type_type_generics, fixed_type_where_clause) =
            fixed_type_generics.split_for_impl();
        let type_struct_ident = format_ident!("__{}__Type", enum_ident);
        let mut field_checks = vec![];
        let mut make_type_struct_variant_type = |variant: &ParsedVariant| {
            let VariantOptions {} = variant.options.body;
            let (value_struct, parsed_struct) = match &variant.value {
                VariantValue::None => {
                    return None;
                }
                VariantValue::Direct { value_type } => {
                    field_checks.push(quote_spanned! {value_type.span()=>
                        __check_field::<#value_type>();
                    });
                    return Some(parse_quote! { <#value_type as ::fayalite::expr::ToExpr>::Type });
                }
                VariantValue::Struct {
                    value_struct,
                    parsed_struct,
                } => (value_struct, parsed_struct),
            };
            value_struct.to_tokens(tokens);
            parsed_struct.to_tokens(tokens);
            let mut field_names = Vec::from_iter(get_field_names(&value_struct.fields));
            derive_clone_hash_eq_partialeq_for_struct(
                &value_struct.ident,
                &fixed_type_generics,
                &field_names,
            )
            .to_tokens(tokens);
            field_names = Vec::from_iter(
                field_names
                    .into_iter()
                    .zip(parsed_struct.fields.iter())
                    .filter_map(|(member, field)| {
                        field.options.body.skip.is_none().then_some(member)
                    }),
            );
            let field_name_strs =
                Vec::from_iter(field_names.iter().map(|v| v.to_token_stream().to_string()));
            let debug_ident = format!("{enum_ident}::{}", variant.ident);
            let debug_body = match variant.fields_kind {
                FieldsKind::Unit => quote! {
                    f.debug_struct(#debug_ident).finish()
                },
                FieldsKind::Named(_) => quote! {
                    f.debug_struct(#debug_ident)
                        #(.field(#field_name_strs, &self.#field_names))*
                        .finish()
                },
                FieldsKind::Unnamed(_) => quote! {
                    f.debug_tuple(#debug_ident)#(.field(&self.#field_names))*.finish()
                },
            };
            let value_struct_ident = &value_struct.ident;
            quote! {
                #[automatically_derived]
                impl #fixed_type_impl_generics ::fayalite::__std::fmt::Debug
                    for #value_struct_ident #fixed_type_type_generics
                #fixed_type_where_clause
                {
                    fn fmt(
                        &self,
                        f: &mut ::fayalite::__std::fmt::Formatter<'_>,
                    ) -> ::fayalite::__std::fmt::Result {
                        #debug_body
                    }
                }
            }
            .to_tokens(tokens);
            Some(parse_quote! {
                <
                    #value_struct_ident #fixed_type_type_generics
                    as ::fayalite::expr::ToExpr
                >::Type
            })
        };
        let type_struct_variants = Punctuated::from_iter(variants.iter().filter_map(|variant| {
            let VariantOptions {} = variant.options.body;
            Some(Field {
                attrs: vec![],
                vis: vis.clone(),
                mutability: FieldMutability::None,
                ident: Some(variant.ident.clone()),
                colon_token: None, // it will fill in the colon if needed
                ty: make_type_struct_variant_type(variant)?,
            })
        }));
        let type_struct = ItemStruct {
            attrs: vec![
                parse_quote! {#[allow(non_camel_case_types)]},
                parse_quote! {#[allow(non_snake_case)]},
            ],
            vis: vis.clone(),
            struct_token: Token![struct](enum_token.span),
            ident: type_struct_ident,
            generics: fixed_type_generics.clone(),
            fields: Fields::Named(FieldsNamed {
                brace_token: *brace_token,
                named: type_struct_variants,
            }),
            semi_token: None,
        };
        let type_struct_ident = &type_struct.ident;
        let type_struct_debug_ident = format!("{enum_ident}::Type");
        type_struct.to_tokens(tokens);
        let non_empty_variant_names = Vec::from_iter(
            variants
                .iter()
                .filter(|v| !v.value.is_none())
                .map(|v| v.ident.clone()),
        );
        let non_empty_variant_name_strs =
            Vec::from_iter(non_empty_variant_names.iter().map(|v| v.to_string()));
        let debug_type_body = quote! {
            f.debug_struct(#type_struct_debug_ident)
                #(.field(#non_empty_variant_name_strs, &self.#non_empty_variant_names))*
                .finish()
        };
        derive_clone_hash_eq_partialeq_for_struct(
            type_struct_ident,
            &fixed_type_generics,
            &non_empty_variant_names,
        )
        .to_tokens(tokens);
        let variant_names = Vec::from_iter(variants.iter().map(|v| &v.ident));
        let variant_name_strs = Vec::from_iter(variant_names.iter().map(|v| v.to_string()));
        let (variant_field_pats, variant_to_canonical_values): (Vec<_>, Vec<_>) = variants
            .iter()
            .map(|v| {
                let field_names: Vec<_> = v.fields.iter().map(|field| &field.name).collect();
                let var_names: Vec<_> = v.fields.iter().map(|field| field.var_name()).collect();
                let field_pats = quote! {
                    #(#field_names: #var_names,)*
                };
                let to_canonical_value = match &v.value {
                    VariantValue::None => quote! { ::fayalite::__std::option::Option::None },
                    VariantValue::Direct { .. } => {
                        debug_assert_eq!(var_names.len(), 1);
                        quote! {
                            ::fayalite::__std::option::Option::Some(
                                ::fayalite::ty::DynValueTrait::to_canonical_dyn(#(#var_names)*),
                            )
                        }
                    }
                    VariantValue::Struct {
                        value_struct,
                        parsed_struct,
                    } => {
                        let value_struct_ident = &value_struct.ident;
                        let phantom_field_name = &parsed_struct
                            .fields
                            .last()
                            .expect("missing phantom field")
                            .name;
                        let type_generics = fixed_type_type_generics.as_turbofish();
                        quote! {
                            ::fayalite::__std::option::Option::Some(
                                ::fayalite::ty::DynValueTrait::to_canonical_dyn(
                                    &#value_struct_ident #type_generics {
                                        #(#field_names:
                                            ::fayalite::__std::clone::Clone::clone(#var_names),)*
                                        #phantom_field_name: ::fayalite::__std::marker::PhantomData,
                                    },
                                ),
                            )
                        }
                    }
                };
                (field_pats, to_canonical_value)
            })
            .unzip();
        let mut match_enum_variants = Punctuated::new();
        let mut match_enum_debug_arms = vec![];
        let mut match_enum_arms = vec![];
        let mut variant_vars = vec![];
        let mut from_canonical_type_variant_lets = vec![];
        let mut non_empty_variant_vars = vec![];
        let mut enum_type_variants = vec![];
        let mut enum_type_variants_hint = vec![];
        let match_enum_ident = format_ident!("__{}__MatchEnum", enum_ident);
        let mut builder = Builder::new(format_ident!("__{}__Builder", enum_ident), vis.clone());
        for variant in variants.iter() {
            for field in variant.fields.iter() {
                builder.insert_field(
                    field.name.clone(),
                    |v| {
                        parse_quote_spanned! {v.span()=>
                            ::fayalite::expr::ToExpr::to_expr(&#v)
                        }
                    },
                    |t| {
                        parse_quote_spanned! {t.span()=>
                            ::fayalite::expr::Expr<<
                                <#t as ::fayalite::expr::ToExpr>::Type
                                as ::fayalite::ty::Type
                            >::Value>
                        }
                    },
                    |t| {
                        parse_quote_spanned! {t.span()=>
                            where
                                #t: ::fayalite::expr::ToExpr,
                        }
                    },
                );
            }
        }
        let builder = builder.finish_filling_in_fields();
        builder.to_tokens(tokens);
        for (variant_index, variant) in variants.iter().enumerate() {
            let variant_var = format_ident!("__v_{}", variant.ident);
            let variant_name = &variant.ident;
            let variant_name_str = variant.ident.to_string();
            match_enum_variants.push(Variant {
                attrs: vec![],
                ident: variant.ident.clone(),
                fields: variant.fields_kind.into_fields(variant.fields.iter().map(
                    |ParsedField {
                         options,
                         vis,
                         name,
                         ty,
                     }| {
                        let FieldOptions {} = options.body;
                        Field {
                            attrs: vec![],
                            vis: vis.clone(),
                            mutability: FieldMutability::None,
                            ident: if let Member::Named(name) = name {
                                Some(name.clone())
                            } else {
                                None
                            },
                            colon_token: None,
                            ty: parse_quote! { ::fayalite::expr::Expr<#ty> },
                        }
                    },
                )),
                discriminant: None,
            });
            let match_enum_field_names = Vec::from_iter(variant.fields.iter().map(
                |ParsedField {
                     options,
                     vis: _,
                     name,
                     ty: _,
                 }| {
                    let FieldOptions {} = options.body;
                    name
                },
            ));
            let match_enum_field_name_strs = Vec::from_iter(variant.fields.iter().map(
                |ParsedField {
                     options,
                     vis: _,
                     name,
                     ty: _,
                 }| {
                    let FieldOptions {} = options.body;
                    name.to_token_stream().to_string()
                },
            ));
            let match_enum_debug_vars = Vec::from_iter(variant.fields.iter().map(
                |ParsedField {
                     options,
                     vis: _,
                     name,
                     ty: _,
                 }| {
                    let FieldOptions {} = options.body;
                    format_ident!("__v_{}", name)
                },
            ));
            match_enum_debug_arms.push(match variant.fields_kind {
                FieldsKind::Unit | FieldsKind::Named(_) => quote! {
                    Self::#variant_name {
                        #(#match_enum_field_names: ref #match_enum_debug_vars,)*
                    } => f.debug_struct(#variant_name_str)
                        #(.field(#match_enum_field_name_strs, #match_enum_debug_vars))*
                        .finish(),
                },
                FieldsKind::Unnamed(_) => quote! {
                    Self::#variant_name(
                        #(ref #match_enum_debug_vars,)*
                    ) => f.debug_tuple(#variant_name_str)
                        #(.field(#match_enum_debug_vars))*
                        .finish(),
                },
            });
            if let Some(value_ty) = variant.value.value_ty() {
                from_canonical_type_variant_lets.push(quote! {
                    let #variant_var =
                        #variant_var.from_canonical_type_helper_has_value(#variant_name_str);
                });
                non_empty_variant_vars.push(variant_var.clone());
                enum_type_variants.push(quote! {
                    ::fayalite::enum_::VariantType {
                        name: ::fayalite::intern::Intern::intern(#variant_name_str),
                        ty: ::fayalite::__std::option::Option::Some(
                            ::fayalite::ty::DynType::canonical_dyn(&self.#variant_name),
                        ),
                    }
                });
                enum_type_variants_hint.push(quote! {
                    ::fayalite::enum_::VariantType {
                        name: ::fayalite::intern::Intern::intern(#variant_name_str),
                        ty: ::fayalite::__std::option::Option::Some(
                            ::fayalite::bundle::TypeHint::<
                                <#value_ty as ::fayalite::expr::ToExpr>::Type,
                            >::intern_dyn(),
                        ),
                    }
                });
            } else {
                from_canonical_type_variant_lets.push(quote! {
                    #variant_var.from_canonical_type_helper_no_value(#variant_name_str);
                });
                enum_type_variants.push(quote! {
                    ::fayalite::enum_::VariantType {
                        name: ::fayalite::intern::Intern::intern(#variant_name_str),
                        ty: ::fayalite::__std::option::Option::None,
                    }
                });
                enum_type_variants_hint.push(quote! {
                    ::fayalite::enum_::VariantType {
                        name: ::fayalite::intern::Intern::intern(#variant_name_str),
                        ty: ::fayalite::__std::option::Option::None,
                    }
                });
            }
            variant_vars.push(variant_var);
            match_enum_arms.push(match &variant.value {
                VariantValue::None => quote! {
                    #variant_index => #match_enum_ident::#variant_name,
                },
                VariantValue::Direct { value_type } => quote! {
                    #variant_index => #match_enum_ident::#variant_name {
                        #(#match_enum_field_names)*: ::fayalite::expr::ToExpr::to_expr(
                            &__variant_access.downcast_unchecked::<
                                <#value_type as ::fayalite::expr::ToExpr>::Type>(),
                        ),
                    },
                },
                VariantValue::Struct {
                    value_struct: ItemStruct { ident, .. },
                    ..
                } => quote! {
                    #variant_index => {
                        let __variant_access = ::fayalite::expr::ToExpr::to_expr(
                            &__variant_access.downcast_unchecked::<<
                                #ident #fixed_type_type_generics
                                as ::fayalite::expr::ToExpr
                            >::Type>(),
                        );
                        #match_enum_ident::#variant_name {
                            #(#match_enum_field_names:
                                (*__variant_access).#match_enum_field_names,)*
                        }
                    },
                },
            });
            let builder_field_and_types = Vec::from_iter(variant.fields.iter().map(
                |ParsedField {
                     options,
                     vis: _,
                     name,
                     ty,
                 }| {
                    let FieldOptions {} = options.body;
                    (name, ty)
                },
            ));
            let builder_field_vars = Vec::from_iter(
                builder_field_and_types
                    .iter()
                    .map(|(name, _)| &builder.get_field(name).unwrap().1.builder_field_name),
            );
            let build_body = match &variant.value {
                VariantValue::None => parse_quote! {
                    {
                        ::fayalite::expr::ToExpr::to_expr(
                            &::fayalite::expr::ops::EnumLiteral::<
                                #type_struct_ident #fixed_type_type_generics
                            >::new_unchecked(
                                ::fayalite::__std::option::Option::None,
                                #variant_index,
                                ::fayalite::ty::FixedType::fixed_type(),
                            ),
                        )
                    }
                },
                VariantValue::Direct { value_type: _ } => parse_quote! {
                    {
                        ::fayalite::expr::ToExpr::to_expr(
                            &::fayalite::expr::ops::EnumLiteral::<
                                #type_struct_ident #fixed_type_type_generics
                            >::new_unchecked(
                                ::fayalite::__std::option::Option::Some(
                                    #(#builder_field_vars)*.to_canonical_dyn(),
                                ),
                                #variant_index,
                                ::fayalite::ty::FixedType::fixed_type(),
                            ),
                        )
                    }
                },
                VariantValue::Struct {
                    parsed_struct:
                        ParsedStruct {
                            names:
                                ParsedStructNames {
                                    type_struct_ident: field_type_struct_ident,
                                    ..
                                },
                            ..
                        },
                    ..
                } => parse_quote! {
                    {
                        let __builder = <
                            #field_type_struct_ident #fixed_type_type_generics
                            as ::fayalite::bundle::BundleType
                        >::builder();
                        #(let __builder = __builder.#builder_field_vars(#builder_field_vars);)*
                        ::fayalite::expr::ToExpr::to_expr(
                            &::fayalite::expr::ops::EnumLiteral::<
                                #type_struct_ident #fixed_type_type_generics
                            >::new_unchecked(
                                ::fayalite::__std::option::Option::Some(
                                    __builder.build().to_canonical_dyn(),
                                ),
                                #variant_index,
                                ::fayalite::ty::FixedType::fixed_type(),
                            ),
                        )
                    }
                },
            };
            builder
                .make_build_method(
                    &format_ident!("variant_{}", variant_name),
                    variant.fields.iter().map(
                        |ParsedField {
                             options,
                             vis: _,
                             name,
                             ty,
                         }| {
                            let FieldOptions {} = options.body;
                            (name.clone(), parse_quote! { ::fayalite::expr::Expr<#ty> })
                        },
                    ),
                    &fixed_type_generics,
                    &parse_quote! {#type_struct_ident #fixed_type_type_generics},
                    &parse_quote! { ::fayalite::expr::Expr<#target #fixed_type_type_generics> },
                    build_body,
                )
                .to_tokens(tokens);
        }
        let match_enum = ItemEnum {
            attrs: vec![parse_quote! {#[allow(non_camel_case_types)]}],
            vis: vis.clone(),
            enum_token: *enum_token,
            ident: match_enum_ident,
            generics: fixed_type_generics.clone(),
            brace_token: *brace_token,
            variants: match_enum_variants,
        };
        let match_enum_ident = &match_enum.ident;
        match_enum.to_tokens(tokens);
        make_connect_impl(
            *connect_inexact,
            &fixed_type_generics,
            type_struct_ident,
            variants.iter().flat_map(|variant| {
                variant.fields.iter().map(|field| {
                    let ty = &field.ty;
                    parse_quote_spanned! {field.name.span()=>
                        <#ty as ::fayalite::expr::ToExpr>::Type
                    }
                })
            }),
        )
        .to_tokens(tokens);
        let variant_count = variants.len();
        let empty_builder_ty = builder.ty([], Some(&parse_quote! { Self }), false);
        quote! {
            #[automatically_derived]
            impl #fixed_type_impl_generics ::fayalite::__std::fmt::Debug
                for #match_enum_ident #fixed_type_type_generics
            #fixed_type_where_clause
            {
                fn fmt(
                    &self,
                    f: &mut ::fayalite::__std::fmt::Formatter<'_>,
                ) -> ::fayalite::__std::fmt::Result {
                    match *self {
                        #(#match_enum_debug_arms)*
                    }
                }
            }

            #[automatically_derived]
            impl #fixed_type_impl_generics ::fayalite::ty::FixedType
                for #type_struct_ident #fixed_type_type_generics
            #fixed_type_where_clause
            {
                fn fixed_type() -> Self {
                    Self {
                        #(#non_empty_variant_names: ::fayalite::ty::FixedType::fixed_type(),)*
                    }
                }
            }

            fn __check_field<T: ::fayalite::ty::Value>()
            where
                <T as ::fayalite::expr::ToExpr>::Type: ::fayalite::ty::Type<Value = T>,
            {}
            fn __check_fields #fixed_type_impl_generics(_: #target #fixed_type_type_generics)
            #fixed_type_where_clause
            {
                #(#field_checks)*
            }

            #[automatically_derived]
            impl #fixed_type_impl_generics ::fayalite::__std::fmt::Debug
                for #type_struct_ident #fixed_type_type_generics
            #fixed_type_where_clause
            {
                fn fmt(
                    &self,
                    f: &mut ::fayalite::__std::fmt::Formatter<'_>,
                ) -> ::fayalite::__std::fmt::Result {
                    #debug_type_body
                }
            }

            #[automatically_derived]
            impl #fixed_type_impl_generics ::fayalite::ty::Type
                for #type_struct_ident #fixed_type_type_generics
            #fixed_type_where_clause
            {
                type CanonicalType = ::fayalite::enum_::DynEnumType;
                type Value = #target #fixed_type_type_generics;
                type CanonicalValue = ::fayalite::enum_::DynEnum;
                type MaskType = ::fayalite::int::UIntType<1>;
                type MaskValue = ::fayalite::int::UInt<1>;
                type MatchVariant = #match_enum_ident #fixed_type_type_generics;
                type MatchActiveScope = ::fayalite::module::Scope;
                type MatchVariantAndInactiveScope =
                    ::fayalite::enum_::EnumMatchVariantAndInactiveScope<Self>;
                type MatchVariantsIter = ::fayalite::enum_::EnumMatchVariantsIter<Self>;
                fn match_variants<IO: ::fayalite::bundle::BundleValue>(
                    this: ::fayalite::expr::Expr<<Self as ::fayalite::ty::Type>::Value>,
                    module_builder: &mut ::fayalite::module::ModuleBuilder<
                        IO,
                        ::fayalite::module::NormalModule,
                    >,
                    source_location: ::fayalite::source_location::SourceLocation,
                ) -> <Self as ::fayalite::ty::Type>::MatchVariantsIter
                where
                    <IO as ::fayalite::expr::ToExpr>::Type:
                        ::fayalite::bundle::BundleType<Value = IO>,
                {
                    module_builder.enum_match_variants_helper(this, source_location)
                }
                fn mask_type(&self) -> <Self as ::fayalite::ty::Type>::MaskType {
                    ::fayalite::int::UIntType::new()
                }
                fn canonical(&self) -> <Self as ::fayalite::ty::Type>::CanonicalType {
                    let variants = ::fayalite::enum_::EnumType::variants(self);
                    ::fayalite::enum_::DynEnumType::new(variants)
                }
                fn source_location(&self) -> ::fayalite::source_location::SourceLocation {
                    ::fayalite::source_location::SourceLocation::caller()
                }
                fn type_enum(&self) -> ::fayalite::ty::TypeEnum {
                    ::fayalite::ty::TypeEnum::EnumType(::fayalite::ty::Type::canonical(self))
                }
                #[allow(non_snake_case)]
                fn from_canonical_type(t: <Self as ::fayalite::ty::Type>::CanonicalType) -> Self {
                    let [#(#variant_vars),*] = *::fayalite::enum_::EnumType::variants(&t) else {
                        ::fayalite::__std::panic!("wrong number of variants");
                    };
                    #(#from_canonical_type_variant_lets)*
                    Self {
                        #(#non_empty_variant_names: #non_empty_variant_vars,)*
                    }
                }
            }

            #[automatically_derived]
            #[allow(clippy::init_numbered_fields)]
            impl #fixed_type_impl_generics ::fayalite::enum_::EnumType
                for #type_struct_ident #fixed_type_type_generics
            #fixed_type_where_clause
            {
                type Builder = #empty_builder_ty;
                fn match_activate_scope(
                    v: <Self as ::fayalite::ty::Type>::MatchVariantAndInactiveScope,
                ) -> (
                    <Self as ::fayalite::ty::Type>::MatchVariant,
                    <Self as ::fayalite::ty::Type>::MatchActiveScope,
                ) {
                    let (__variant_access, __scope) = v.activate();
                    (
                        match ::fayalite::expr::ops::VariantAccess::variant_index(
                            &*__variant_access,
                        ) {
                            #(#match_enum_arms)*
                            #variant_count.. => ::fayalite::__std::panic!("invalid variant index"),
                        },
                        __scope,
                    )
                }
                fn builder() -> <Self as ::fayalite::enum_::EnumType>::Builder {
                    #empty_builder_ty::new()
                }
                fn variants(&self) -> ::fayalite::intern::Interned<[::fayalite::enum_::VariantType<
                    ::fayalite::intern::Interned<dyn ::fayalite::ty::DynCanonicalType>,
                >]> {
                    ::fayalite::intern::Intern::intern(&[#(#enum_type_variants,)*][..])
                }
                fn variants_hint() -> ::fayalite::enum_::VariantsHint {
                    ::fayalite::enum_::VariantsHint::new([#(#enum_type_variants_hint,)*], false)
                }
            }

            #[automatically_derived]
            impl #fixed_type_impl_generics ::fayalite::expr::ToExpr
                for #target #fixed_type_type_generics
            #fixed_type_where_clause
            {
                type Type = #type_struct_ident #fixed_type_type_generics;
                fn ty(&self) -> <Self as ::fayalite::expr::ToExpr>::Type {
                    ::fayalite::ty::FixedType::fixed_type()
                }
                fn to_expr(&self) -> ::fayalite::expr::Expr<Self> {
                    ::fayalite::expr::Expr::from_value(self)
                }
            }

            #[automatically_derived]
            impl #fixed_type_impl_generics ::fayalite::ty::Value
                for #target #fixed_type_type_generics
            #fixed_type_where_clause
            {
                fn to_canonical(&self) -> <
                        <Self as ::fayalite::expr::ToExpr>::Type
                        as ::fayalite::ty::Type
                    >::CanonicalValue
                {
                    let __ty = ::fayalite::ty::Type::canonical(&::fayalite::expr::ToExpr::ty(self));
                    match self {
                        #(Self::#variant_names { #variant_field_pats } => {
                            ::fayalite::enum_::DynEnum::new_by_name(
                                __ty,
                                ::fayalite::intern::Intern::intern(#variant_name_strs),
                                #variant_to_canonical_values,
                            )
                        })*
                    }
                }
            }

            #[automatically_derived]
            impl #fixed_type_impl_generics ::fayalite::enum_::EnumValue
                for #target #fixed_type_type_generics
            #fixed_type_where_clause
            {
            }
        }
        .to_tokens(tokens);
    }
}

pub(crate) fn value_derive_enum(item: ItemEnum) -> syn::Result<TokenStream> {
    let item = ParsedEnum::parse(item)?;
    let outline_generated = item.options.body.outline_generated;
    let mut contents = quote! {
        const _: () = {
            #item
        };
    };
    if outline_generated.is_some() {
        contents = crate::outline_generated(contents, "value-enum-");
    }
    Ok(contents)
}
