//! # Memories
//!
//! Memories are optimized for storing large amounts of data.
//!
//! When you create a memory, you get a [`MemBuilder`], which you
//! can then use to add memory ports, which is how you can read/write the memory.
//!
//! There are several different ways to create a memory:
//!
//! ## using [`ModuleBuilder::memory()`]
//!
//! This way you have to set the [`depth`][`MemBuilder::depth`] separately.
//!
//! ```
//! # use fayalite::{hdl_module, int::UInt, clock::ClockDomain};
//! # #[hdl_module]
//! # fn module() {
//! // first, we need some IO
//! #[hdl]
//! let cd: ClockDomain = m.input();
//! #[hdl]
//! let read_addr: UInt<8> = m.input();
//! #[hdl]
//! let read_data: UInt<8> = m.output();
//!
//! // now create the memory
//! #[hdl]
//! let mut my_memory = m.memory();
//! my_memory.depth(256); // the memory has 256 elements
//!
//! let read_port = my_memory.new_read_port();
//!
//! // connect up the read port
//! m.connect_any(read_port.addr, read_addr);
//! m.connect(read_port.en, 1_hdl_u1);
//! m.connect(read_port.clk, cd.clk);
//! m.connect(read_data, read_port.data);
//!
//! // we need more IO for the write port
//! #[hdl]
//! let write_addr: UInt<8> = m.input();
//! #[hdl]
//! let do_write: UInt<1> = m.input();
//! #[hdl]
//! let write_data: UInt<8> = m.input();
//!
//! let write_port = my_memory.new_write_port();
//!
//! m.connect_any(write_port.addr, write_addr);
//! m.connect(write_port.en, do_write);
//! m.connect(write_port.clk, cd.clk);
//! m.connect(write_port.data, write_port.data);
//! m.connect(write_port.mask, 1_hdl_u1);
//! # }
//! ```
//!
//! ## using [`ModuleBuilder::memory_array()`]
//!
//! this allows you to specify the memory's underlying array type directly.
//!
//! ```
//! # use fayalite::{hdl_module, int::UInt, memory::MemBuilder};
//! # #[hdl_module]
//! # fn module() {
//! #[hdl]
//! let mut my_memory: MemBuilder<[UInt<8>; 256]> = m.memory_array();
//!
//! let read_port = my_memory.new_read_port();
//! // ...
//! let write_port = my_memory.new_write_port();
//! // ...
//! # }
//! ```
//!
//! ## using [`ModuleBuilder::memory_with_init()`]
//!
//! This allows you to deduce the memory's array type from the data used to initialize the memory.
//!
//! ```
//! # use fayalite::{hdl_module, int::UInt};
//! # #[hdl_module]
//! # fn module() {
//! # #[hdl]
//! # let read_addr: UInt<2> = m.input();
//! #[hdl]
//! let mut my_memory = m.memory_with_init(
//!     #[hdl]
//!     [0x12_hdl_u8, 0x34_hdl_u8, 0x56_hdl_u8, 0x78_hdl_u8],
//! );
//!
//! let read_port = my_memory.new_read_port();
//! // note that `read_addr` is `UInt<2>` since the memory only has 4 elements
//! m.connect_any(read_port.addr, read_addr);
//! // ...
//! let write_port = my_memory.new_write_port();
//! // ...
//! # }
//! ```

#[allow(unused)]
use crate::{memory::MemBuilder, module::ModuleBuilder};
