//! # Connection Semantics
//!
//! Fayalite's connection semantics are unlike assignments in software, so be careful!
//!
//! Fayalite's connection semantics follow [FIRRTL]'s Last-Connect-Semantics and
//! Conditional-Last-Connect-Semantics:
//!
//! Basically, every [wire] behaves as if you ran all connections in a module, and everywhere
//! the wire is read from, it takes on the value it has at the end of the module, with every
//! connect (except those in any kind of [conditional block] where the condition doesn't hold,
//! such as an [`#[hdl] if`][if] with a false condition).
//! overwriting the appropriate portion of the wire.
//!
//! Any other things that are connected to (on the LHS of a
//! [`connect()`] or [`connect_any()`] call) have analogous connection semantics.
//!
//! This description is intended to match [FIRRTL]'s description, so if they conflict with
//! each other, please report it as a bug in Fayalite.
//!
//! Connection Semantics Example:
//!
//! ```
//! # use fayalite::module_hdl;
//! # #[module_hdl]
//! # fn module() {
//! #[hdl]
//! let a: UInt<8> = m.wire();
//! #[hdl]
//! let b: UInt<8> = m.output();
//!
//! // doesn't actually affect anything, since `a` is completely overwritten later
//! m.connect(a, 5_hdl_u8);
//!
//! // here `a` has value `7` since the last connection assigns
//! // `7` to `a`, so `b` has value `7` too.
//! m.connect(b, a);
//!
//! // this is the last `connect` to `a`, so this `connect` determines `a`'s value
//! m.connect(a, 7_hdl_u8);
//! # }
//! ```
//!
//! # Conditional Connection Semantics
//!
//! ```
//! # use fayalite::module_hdl;
//! # #[module_hdl]
//! # fn module() {
//! #[hdl]
//! let cond: UInt<1> = m.input();
//! #[hdl]
//! let a: UInt<8> = m.wire();
//! #[hdl]
//! let b: UInt<8> = m.output();
//!
//! // this is the last `connect` to `a` when `cond` is `0`
//! m.connect(a, 5_hdl_u8);
//!
//! // here `a` has value `7` if `cond` is `1` since the last connection assigns
//! // `7` to `a`, so `b` has value `7` too, otherwise `a` (and therefore `b`)
//! // have value `5` since then the connection assigning `7` is in a
//! // conditional block where the condition doesn't hold.
//! m.connect(b, a);
//!
//! #[hdl]
//! if cond {
//!     // this is the last `connect` to `a` when `cond` is `1`
//!     m.connect(a, 7_hdl_u8);
//! }
//! # }
//! ```
//!
//! [conditional block]: self#conditional-connection-semantics
//! [`connect()`]: ModuleBuilder::connect
//! [`connect_any()`]: ModuleBuilder::connect_any
//! [wire]: crate::_docs::modules::module_bodies::hdl_let_statements::wires
//! [if]: crate::_docs::modules::module_bodies::hdl_if_statements
//! [FIRRTL]: https://github.com/chipsalliance/firrtl-spec

#[allow(unused)]
use crate::module::ModuleBuilder;
