// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    array::{Array, ArrayType},
    bundle::{BundleType, BundleValue, DynBundle, DynBundleType},
    enum_::{DynEnum, DynEnumType, EnumType, EnumValue, VariantType},
    expr::{ops, Expr, ExprEnum, ToExpr},
    int::{DynUInt, DynUIntType, IntCmp},
    intern::{Intern, Interned},
    memory::{DynPortType, Mem, MemPort},
    module::{
        transform::visit::{Fold, Folder},
        Block, Module, NameIdGen, ScopedNameId, Stmt, StmtConnect, StmtIf, StmtMatch, StmtWire,
    },
    source_location::SourceLocation,
    ty::{DynCanonicalType, DynCanonicalValue, Type, TypeEnum, Value, ValueEnum},
    wire::Wire,
};
use core::fmt;
use hashbrown::HashMap;

#[derive(Debug)]
pub enum SimplifyEnumsError {
    EnumIsNotCastableFromBits { enum_type: DynEnumType },
}

impl fmt::Display for SimplifyEnumsError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            SimplifyEnumsError::EnumIsNotCastableFromBits { enum_type } => write!(
                f,
                "simplify_enums failed: enum type is not castable from bits: {enum_type:?}"
            ),
        }
    }
}

impl std::error::Error for SimplifyEnumsError {}

#[derive(Value, Clone, Eq, PartialEq, Hash, Debug)]
struct TagAndBody<T> {
    tag: T,
    body: DynUInt,
}

type TagAndBodyType<T> = <TagAndBody<T> as ToExpr>::Type;

#[derive(Clone, Debug)]
enum EnumTypeState {
    TagEnumAndBody(TagAndBodyType<DynEnum>),
    TagUIntAndBody(TagAndBodyType<DynUInt>),
    UInt(DynUIntType),
    Unchanged,
}

struct State {
    enum_types: HashMap<DynEnumType, EnumTypeState>,
    replacement_mem_ports:
        HashMap<Interned<MemPort<DynPortType>>, Interned<Wire<Interned<dyn DynCanonicalType>>>>,
    kind: SimplifyEnumsKind,
    name_id_gen: NameIdGen,
}

impl State {
    fn get_or_make_enum_type_state(
        &mut self,
        enum_type: DynEnumType,
    ) -> Result<EnumTypeState, SimplifyEnumsError> {
        if let Some(retval) = self.enum_types.get(&enum_type) {
            return Ok(retval.clone());
        }
        if !enum_type.is_castable_from_bits() {
            return Err(SimplifyEnumsError::EnumIsNotCastableFromBits { enum_type });
        }
        let has_body = enum_type
            .variants()
            .iter()
            .any(|variant| variant.ty.is_some());
        let retval = match (self.kind, has_body) {
            (SimplifyEnumsKind::SimplifyToEnumsWithNoBody, true) => {
                EnumTypeState::TagEnumAndBody(TagAndBodyType::<DynEnum> {
                    tag: DynEnumType::new(Interned::from_iter(enum_type.variants().iter().map(
                        |v| VariantType {
                            name: v.name,
                            ty: None,
                        },
                    ))),
                    body: DynUIntType::new(
                        enum_type.bit_width() - enum_type.discriminant_bit_width(),
                    ),
                })
            }
            (SimplifyEnumsKind::SimplifyToEnumsWithNoBody, false) => EnumTypeState::Unchanged,
            (SimplifyEnumsKind::ReplaceWithBundleOfUInts, _) => {
                EnumTypeState::TagUIntAndBody(TagAndBodyType::<DynUInt> {
                    tag: DynUIntType::new(enum_type.discriminant_bit_width()),
                    body: DynUIntType::new(
                        enum_type.bit_width() - enum_type.discriminant_bit_width(),
                    ),
                })
            }
            (SimplifyEnumsKind::ReplaceWithUInt, _) => {
                EnumTypeState::UInt(DynUIntType::new(enum_type.bit_width()))
            }
        };
        self.enum_types.insert(enum_type, retval.clone());
        Ok(retval)
    }
}

fn value_to_uint<T: Value>(value: Option<&T>, target_ty: DynUIntType) -> DynUInt {
    let Some(value) = value else {
        return DynUInt::with_type(target_ty, 0u8);
    };
    DynUInt::from_bit_slice(&value.to_bits())
}

fn connect_port(
    stmts: &mut Vec<Stmt>,
    lhs: Expr<DynCanonicalValue>,
    rhs: Expr<DynCanonicalValue>,
    source_location: SourceLocation,
) {
    println!("connect_port: lhs={lhs:?} rhs={rhs:?}");
    if lhs.canonical_type() == rhs.canonical_type() {
        stmts.push(
            dbg!(StmtConnect {
                lhs,
                rhs,
                source_location,
            })
            .into(),
        );
        return;
    }
    match (
        lhs.canonical_type().type_enum(),
        rhs.canonical_type().type_enum(),
    ) {
        (TypeEnum::BundleType(lhs_type), TypeEnum::UInt(_)) => {
            let lhs = lhs.with_type::<DynBundle>();
            for field in lhs_type.fields() {
                assert!(!field.flipped);
                connect_port(stmts, lhs.field(&field.name), rhs, source_location);
            }
        }
        (TypeEnum::UInt(_), TypeEnum::BundleType(rhs_type)) => {
            let rhs = rhs.with_type::<DynBundle>();
            for field in rhs_type.fields() {
                assert!(!field.flipped);
                connect_port(stmts, lhs, rhs.field(&field.name), source_location);
            }
        }
        (TypeEnum::BundleType(lhs_type), TypeEnum::BundleType(_)) => {
            let lhs = lhs.with_type::<DynBundle>();
            let rhs = rhs.with_type::<DynBundle>();
            for field in lhs_type.fields() {
                let (lhs_field, rhs_field) = if field.flipped {
                    (rhs.field(&field.name), lhs.field(&field.name))
                } else {
                    (lhs.field(&field.name), rhs.field(&field.name))
                };
                connect_port(stmts, lhs_field, rhs_field, source_location);
            }
        }
        (TypeEnum::ArrayType(lhs_type), TypeEnum::ArrayType(_)) => {
            let lhs = lhs.with_type::<Array<[DynCanonicalValue]>>();
            let rhs = rhs.with_type::<Array<[DynCanonicalValue]>>();
            for index in 0..lhs_type.len() {
                connect_port(stmts, lhs[index], rhs[index], source_location);
            }
        }
        (TypeEnum::BundleType(_), _)
        | (TypeEnum::EnumType(_), _)
        | (TypeEnum::ArrayType(_), _)
        | (TypeEnum::UInt(_), _)
        | (TypeEnum::SInt(_), _)
        | (TypeEnum::Clock(_), _)
        | (TypeEnum::AsyncReset(_), _)
        | (TypeEnum::SyncReset(_), _)
        | (TypeEnum::Reset(_), _) => unreachable!(
            "trying to connect memory ports:\n{:?}\n{:?}",
            lhs.canonical_type().type_enum(),
            rhs.canonical_type().type_enum(),
        ),
    }
}

impl Folder for State {
    type Error = SimplifyEnumsError;

    fn fold_dyn_enum(&mut self, _v: DynEnum) -> Result<DynEnum, Self::Error> {
        unreachable!()
    }

    fn fold_dyn_enum_type(&mut self, _v: DynEnumType) -> Result<DynEnumType, Self::Error> {
        unreachable!()
    }

    fn fold_module<T: BundleValue>(&mut self, v: Module<T>) -> Result<Module<T>, Self::Error>
    where
        T::Type: BundleType<Value = T>,
    {
        let old_name_id_gen =
            std::mem::replace(&mut self.name_id_gen, NameIdGen::for_module(v.canonical()));
        let retval = Fold::default_fold(v, self);
        self.name_id_gen = old_name_id_gen;
        retval
    }

    fn fold_expr_enum(&mut self, op: ExprEnum) -> Result<ExprEnum, Self::Error> {
        match op {
            ExprEnum::EnumLiteral(op) => Ok(match self.get_or_make_enum_type_state(op.ty())? {
                EnumTypeState::TagEnumAndBody(TagAndBodyType::<DynEnum> { tag, body }) => {
                    TagAndBodyType::<DynEnum>::builder()
                        .field_tag(DynEnum::new_by_index(tag, op.variant_index(), None))
                        .field_body(match op.variant_value() {
                            Some(variant_value) => variant_value.fold(self)?.cast_to_bits(),
                            None => DynUInt::with_type(body, 0u8).to_expr(),
                        })
                        .build()
                        .expr_enum()
                }
                EnumTypeState::TagUIntAndBody(TagAndBodyType::<DynUInt> { tag, body }) => {
                    TagAndBodyType::<DynUInt>::builder()
                        .field_tag(DynUInt::with_type(tag, op.variant_index()))
                        .field_body(match op.variant_value() {
                            Some(variant_value) => variant_value.fold(self)?.cast_to_bits(),
                            None => DynUInt::with_type(body, 0u8).to_expr(),
                        })
                        .build()
                        .expr_enum()
                }
                EnumTypeState::UInt(_) => TagAndBodyType::<DynUInt>::builder()
                    .field_tag(DynUInt::with_type(
                        DynUIntType::new(op.ty().discriminant_bit_width()),
                        op.variant_index(),
                    ))
                    .field_body(match op.variant_value() {
                        Some(variant_value) => variant_value.fold(self)?.cast_to_bits(),
                        None => DynUInt::with_type(
                            DynUIntType::new(
                                op.ty().bit_width() - op.ty().discriminant_bit_width(),
                            ),
                            0u8,
                        )
                        .to_expr(),
                    })
                    .build()
                    .cast_to_bits()
                    .expr_enum(),
                EnumTypeState::Unchanged => ExprEnum::EnumLiteral(
                    ops::EnumLiteral::new_unchecked(
                        op.variant_value().map(|v| v.fold(self)).transpose()?,
                        op.variant_index(),
                        op.ty(),
                    )
                    .intern_sized(),
                ),
            }),
            ExprEnum::VariantAccess(op) => {
                Ok(match self.get_or_make_enum_type_state(op.base().ty())? {
                    EnumTypeState::TagEnumAndBody(_) | EnumTypeState::TagUIntAndBody(_) => {
                        match op.variant_type().ty {
                            Some(_) => op
                                .base()
                                .expr_enum()
                                .fold(self)?
                                .to_expr()
                                .with_type::<TagAndBody<DynCanonicalValue>>()
                                .body[..op.ty().bit_width()]
                                .cast_bits_to::<DynCanonicalValue>(op.ty())
                                .expr_enum(),
                            None => ().to_expr().expr_enum(),
                        }
                    }
                    EnumTypeState::UInt(_) => match op.variant_type().ty {
                        Some(_) => {
                            let base_int = op
                                .base()
                                .expr_enum()
                                .fold(self)?
                                .to_expr()
                                .with_type::<DynUInt>();
                            dbg!(base_int);
                            let base_ty = op.base().ty();
                            let ty_bit_width = op.ty().bit_width();
                            base_int[base_ty.discriminant_bit_width()..][..ty_bit_width]
                                .cast_bits_to::<DynCanonicalValue>(op.ty())
                                .expr_enum()
                        }
                        None => ().to_expr().expr_enum(),
                    },
                    EnumTypeState::Unchanged => match op.variant_type().ty {
                        Some(_) => ExprEnum::VariantAccess(
                            ops::VariantAccess::new_unchecked(
                                op.base().fold(self)?,
                                op.variant_index(),
                            )
                            .intern_sized(),
                        ),
                        None => ().to_expr().expr_enum(),
                    },
                })
            }
            ExprEnum::MemPort(mem_port) => Ok(
                if let Some(&wire) = self.replacement_mem_ports.get(&mem_port) {
                    ExprEnum::Wire(wire)
                } else {
                    ExprEnum::MemPort(mem_port.fold(self)?)
                },
            ),
            ExprEnum::Literal(_)
            | ExprEnum::ArrayLiteral(_)
            | ExprEnum::BundleLiteral(_)
            | ExprEnum::NotU(_)
            | ExprEnum::NotS(_)
            | ExprEnum::Neg(_)
            | ExprEnum::BitAndU(_)
            | ExprEnum::BitAndS(_)
            | ExprEnum::BitOrU(_)
            | ExprEnum::BitOrS(_)
            | ExprEnum::BitXorU(_)
            | ExprEnum::BitXorS(_)
            | ExprEnum::AddU(_)
            | ExprEnum::AddS(_)
            | ExprEnum::SubU(_)
            | ExprEnum::SubS(_)
            | ExprEnum::MulU(_)
            | ExprEnum::MulS(_)
            | ExprEnum::DynShlU(_)
            | ExprEnum::DynShlS(_)
            | ExprEnum::DynShrU(_)
            | ExprEnum::DynShrS(_)
            | ExprEnum::FixedShlU(_)
            | ExprEnum::FixedShlS(_)
            | ExprEnum::FixedShrU(_)
            | ExprEnum::FixedShrS(_)
            | ExprEnum::CmpLtU(_)
            | ExprEnum::CmpLtS(_)
            | ExprEnum::CmpLeU(_)
            | ExprEnum::CmpLeS(_)
            | ExprEnum::CmpGtU(_)
            | ExprEnum::CmpGtS(_)
            | ExprEnum::CmpGeU(_)
            | ExprEnum::CmpGeS(_)
            | ExprEnum::CmpEqU(_)
            | ExprEnum::CmpEqS(_)
            | ExprEnum::CmpNeU(_)
            | ExprEnum::CmpNeS(_)
            | ExprEnum::CastUIntToUInt(_)
            | ExprEnum::CastUIntToSInt(_)
            | ExprEnum::CastSIntToUInt(_)
            | ExprEnum::CastSIntToSInt(_)
            | ExprEnum::SliceUInt(_)
            | ExprEnum::SliceSInt(_)
            | ExprEnum::ReduceBitAnd(_)
            | ExprEnum::ReduceBitOr(_)
            | ExprEnum::ReduceBitXor(_)
            | ExprEnum::FieldAccess(_)
            | ExprEnum::ArrayIndex(_)
            | ExprEnum::DynArrayIndex(_)
            | ExprEnum::CastToBits(_)
            | ExprEnum::CastBitsTo(_)
            | ExprEnum::CastBitToClock(_)
            | ExprEnum::CastBitToSyncReset(_)
            | ExprEnum::CastBitToAsyncReset(_)
            | ExprEnum::CastSyncResetToReset(_)
            | ExprEnum::CastAsyncResetToReset(_)
            | ExprEnum::CastClockToBit(_)
            | ExprEnum::CastSyncResetToBit(_)
            | ExprEnum::CastAsyncResetToBit(_)
            | ExprEnum::CastResetToBit(_)
            | ExprEnum::ModuleIO(_)
            | ExprEnum::Instance(_)
            | ExprEnum::Wire(_)
            | ExprEnum::Reg(_) => op.default_fold(self),
        }
    }

    fn fold_block(&mut self, block: Block) -> Result<Block, Self::Error> {
        let mut memories = vec![];
        let mut stmts = vec![];
        for memory in block.memories {
            let old_element_ty = *memory.array_type().element();
            let new_element_ty = memory.array_type().element().fold(self)?;
            if new_element_ty != old_element_ty {
                let mut new_ports = vec![];
                for port in memory.ports() {
                    let new_port = MemPort::<DynPortType>::new_unchecked(
                        port.mem_name(),
                        port.source_location(),
                        port.port_name(),
                        port.addr_type(),
                        new_element_ty,
                    );
                    new_ports.push(new_port.intern());
                    let new_port_ty = new_port.ty();
                    let mut wire_ty_fields = Vec::from_iter(new_port_ty.fields());
                    if let Some(wmask_name) = new_port.port_kind().wmask_name() {
                        let index = *new_port_ty
                            .name_indexes()
                            .get(&wmask_name.intern())
                            .unwrap();
                        wire_ty_fields[index].ty = port.ty().fields()[index].ty;
                    }
                    let wire_ty = DynBundleType::new(Intern::intern_owned(wire_ty_fields));
                    if wire_ty == new_port_ty {
                        continue;
                    }
                    let wire_name = self.name_id_gen.gen(
                        (*format!("{}_{}", memory.scoped_name().1 .0, port.port_name())).intern(),
                    );
                    let wire = Wire::new_unchecked(
                        ScopedNameId(memory.scoped_name().0, wire_name),
                        port.source_location(),
                        wire_ty,
                    );
                    stmts.push(
                        StmtWire {
                            annotations: Default::default(),
                            wire: wire.to_dyn_canonical_wire(),
                        }
                        .into(),
                    );
                    connect_port(
                        &mut stmts,
                        new_port.to_expr().to_canonical_dyn(),
                        wire.to_expr().to_canonical_dyn(),
                        port.source_location(),
                    );
                    self.replacement_mem_ports
                        .insert(port, wire.to_dyn_canonical_wire().intern_sized());
                }
                memories.push(Mem::new_unchecked(
                    memory.scoped_name(),
                    memory.source_location(),
                    ArrayType::new_slice(new_element_ty, memory.array_type().len()),
                    memory.initial_value(),
                    Intern::intern_owned(new_ports),
                    memory.read_latency(),
                    memory.write_latency(),
                    memory.read_under_write(),
                    memory.port_annotations(),
                    memory.mem_annotations(),
                ));
            } else {
                memories.push(memory.fold(self)?);
            }
        }
        stmts.extend_from_slice(&block.stmts.fold(self)?);
        Ok(Block {
            memories: Intern::intern_owned(memories),
            stmts: Intern::intern_owned(stmts),
        })
    }

    fn fold_stmt(&mut self, stmt: Stmt) -> Result<Stmt, Self::Error> {
        fn match_int_tag(
            state: &mut State,
            int_tag_expr: Expr<DynUInt>,
            source_location: SourceLocation,
            blocks: Interned<[Block]>,
        ) -> Result<StmtIf, SimplifyEnumsError> {
            let mut blocks_iter = blocks.iter().copied().enumerate();
            let (_, last_block) = blocks_iter.next_back().unwrap_or_default();
            let Some((next_to_last_variant_index, next_to_last_block)) = blocks_iter.next_back()
            else {
                return Ok(StmtIf {
                    cond: true.to_expr(),
                    source_location,
                    blocks: [last_block.fold(state)?, Block::default()],
                });
            };
            let mut retval = StmtIf {
                cond: int_tag_expr.cmp_eq(DynUInt::with_type(
                    int_tag_expr.ty(),
                    next_to_last_variant_index,
                )),
                source_location,
                blocks: [next_to_last_block.fold(state)?, last_block.fold(state)?],
            };
            for (variant_index, block) in blocks_iter.rev() {
                retval = StmtIf {
                    cond: int_tag_expr.cmp_eq(DynUInt::with_type(int_tag_expr.ty(), variant_index)),
                    source_location,
                    blocks: [
                        block.fold(state)?,
                        Block {
                            memories: Default::default(),
                            stmts: [Stmt::from(retval)][..].intern(),
                        },
                    ],
                };
            }
            Ok(retval)
        }
        match stmt {
            Stmt::Match(StmtMatch {
                expr,
                source_location,
                blocks,
            }) => match self.get_or_make_enum_type_state(expr.ty())? {
                EnumTypeState::TagEnumAndBody(_) => Ok(StmtMatch {
                    expr: expr
                        .expr_enum()
                        .fold(self)?
                        .to_expr()
                        .with_type::<TagAndBody<DynEnum>>()
                        .tag,
                    source_location,
                    blocks: blocks.fold(self)?,
                }
                .into()),
                EnumTypeState::TagUIntAndBody(_) => {
                    let int_tag_expr = expr
                        .expr_enum()
                        .fold(self)?
                        .to_expr()
                        .with_type::<TagAndBody<DynUInt>>()
                        .tag;
                    Ok(match_int_tag(self, int_tag_expr, source_location, blocks)?.into())
                }
                EnumTypeState::UInt(_) => {
                    let int_tag_expr = expr
                        .expr_enum()
                        .fold(self)?
                        .to_expr()
                        .with_type::<DynUInt>()[..expr.ty().discriminant_bit_width()];
                    Ok(match_int_tag(self, int_tag_expr, source_location, blocks)?.into())
                }
                EnumTypeState::Unchanged => Ok(StmtMatch {
                    expr: expr.fold(self)?,
                    source_location,
                    blocks: blocks.fold(self)?,
                }
                .into()),
            },
            Stmt::Connect(_) | Stmt::If(_) | Stmt::Declaration(_) => stmt.default_fold(self),
        }
    }

    fn fold_stmt_match(&mut self, _v: StmtMatch) -> Result<StmtMatch, Self::Error> {
        unreachable!()
    }

    fn fold_type_enum(&mut self, type_enum: TypeEnum) -> Result<TypeEnum, Self::Error> {
        match type_enum {
            TypeEnum::EnumType(enum_type) => {
                Ok(match self.get_or_make_enum_type_state(enum_type)? {
                    EnumTypeState::TagEnumAndBody(ty) => TypeEnum::BundleType(ty.canonical()),
                    EnumTypeState::TagUIntAndBody(ty) => TypeEnum::BundleType(ty.canonical()),
                    EnumTypeState::UInt(ty) => TypeEnum::UInt(ty),
                    EnumTypeState::Unchanged => TypeEnum::EnumType(enum_type),
                })
            }
            TypeEnum::BundleType(_)
            | TypeEnum::ArrayType(_)
            | TypeEnum::UInt(_)
            | TypeEnum::SInt(_)
            | TypeEnum::Clock(_)
            | TypeEnum::AsyncReset(_)
            | TypeEnum::SyncReset(_)
            | TypeEnum::Reset(_) => type_enum.default_fold(self),
        }
    }

    fn fold_value_enum(&mut self, value_enum: ValueEnum) -> Result<ValueEnum, Self::Error> {
        match value_enum {
            ValueEnum::Enum(enum_value) => {
                Ok(match self.get_or_make_enum_type_state(enum_value.ty())? {
                    EnumTypeState::TagEnumAndBody(TagAndBodyType::<DynEnum> {
                        tag: tag_ty,
                        body: body_ty,
                    }) => ValueEnum::Bundle(
                        TagAndBody {
                            tag: DynEnum::new_by_index(tag_ty, enum_value.variant_index(), None),
                            body: value_to_uint(enum_value.variant_value().as_ref(), body_ty),
                        }
                        .to_canonical(),
                    ),
                    EnumTypeState::TagUIntAndBody(TagAndBodyType::<DynUInt> {
                        tag: tag_ty,
                        body: body_ty,
                    }) => ValueEnum::Bundle(
                        TagAndBody {
                            tag: DynUInt::with_type(tag_ty, enum_value.variant_index()),
                            body: value_to_uint(enum_value.variant_value().as_ref(), body_ty),
                        }
                        .to_canonical(),
                    ),
                    EnumTypeState::UInt(target_ty) => {
                        ValueEnum::UInt(value_to_uint(Some(&enum_value), target_ty))
                    }
                    EnumTypeState::Unchanged => ValueEnum::Enum(enum_value),
                })
            }
            ValueEnum::Bundle(_)
            | ValueEnum::Array(_)
            | ValueEnum::UInt(_)
            | ValueEnum::SInt(_)
            | ValueEnum::Clock(_)
            | ValueEnum::AsyncReset(_)
            | ValueEnum::SyncReset(_)
            | ValueEnum::Reset(_) => value_enum.default_fold(self),
        }
    }

    fn fold_variant_type<T>(&mut self, _v: VariantType<T>) -> Result<VariantType<T>, Self::Error> {
        unreachable!()
    }

    fn fold_enum_literal<EnumTy>(
        &mut self,
        _v: ops::EnumLiteral<EnumTy>,
    ) -> Result<ops::EnumLiteral<EnumTy>, Self::Error>
    where
        EnumTy: EnumType,
        EnumTy::Value: EnumValue<Type = EnumTy>,
    {
        unreachable!()
    }

    fn fold_variant_access<T, VariantTy>(
        &mut self,
        _v: ops::VariantAccess<T, VariantTy>,
    ) -> Result<ops::VariantAccess<T, VariantTy>, Self::Error>
    where
        T: EnumType,
        T::Value: EnumValue,
        VariantTy: Type,
    {
        unreachable!()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub enum SimplifyEnumsKind {
    SimplifyToEnumsWithNoBody,
    ReplaceWithBundleOfUInts,
    ReplaceWithUInt,
}

pub fn simplify_enums(
    module: Interned<Module<DynBundle>>,
    kind: SimplifyEnumsKind,
) -> Result<Interned<Module<DynBundle>>, SimplifyEnumsError> {
    module.fold(&mut State {
        enum_types: HashMap::new(),
        replacement_mem_ports: HashMap::new(),
        kind,
        name_id_gen: NameIdGen::default(),
    })
}
